#############################################################################
# clive, video extraction utility
# Copyright (C) 2007 Toni Gundogdu
#
# clive is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  0.1.2-1307  USA
#############################################################################

import os
import sys
import formatter
import urllib

from cStringIO import StringIO
from htmllib import HTMLParser
from urllib import *

from clive.util import *
from clive.opts import *


__all__ = ['Extract']


class Extract:
  """Core class for extracting videos from websites"""
  def __init__(self, onsay, onprogress, onprompt):
    self._onsay = onsay
    self._onprogress = onprogress
    self._onprompt = onprompt

    o = Options()
    (self.opts, self.args, self.conf) = o.parse()

    self._setupconn()

  def get(self):
    try:
      if not self.opts.batch:
        url = self._checkurl()
      else:
        url = None
      self._get(url)

    except CliveError,e:
      self._say(e.msg)

  # Non-public

  def _say(self, text):
    """Wraps _onsay for --quiet,--nostrip"""
    if self.opts.verbose:
      if self.opts.strip and text != '\n' and text[0] != '-':
        self._onsay(self._striptext(text))
      else:
        self._onsay(text)

  def _striptext(self, text):
    """Strips text to fit a line"""
    text = text[:70]
    if len(text) >= 70:
      text += ' (...)'
    return text

  def _progress(self, blocks, blocksize, total):
    """Wraps _onprogress for --quiet"""
    if self.opts.verbose:
      self._onprogress(blocks, blocksize, total)

  def _setupconn(self):
    """Sets up the connection elements user-agent and non-auth proxy"""
    proxies = {} # Don't use proxy
    if self.opts.proxy:
      if self.opts.proxy == 'env':
        # Use environment variable http_proxy
        proxies = None
      else:
        # Use the designated proxy
        proxies = {'http':self.opts.proxy}

    urllib._urlopener = CliveOpener(self.opts.agent, proxies)

  def _checkurl(self):
    """
    1: Prompt for url if unspecified
    2: Check for URL length
    3: Remove any unwanted chars from the URL and apply basic URL filtering
    """
    if len(self.args) != 1:
      self.args.append(self._onprompt('URL: '))

    # Check URL length
    if len(self.args[0]) == 0:
      raise CliveError('Error: Invalid URL')
  
    # Convert embed to video page URL
    return normalize_url(self.args[0])

  def _printsep(self):
    """Prints out a separator (dash) line"""
    self._say('-'*70)

  def _get(self, url):
    self._say('Agent: %s' % urllib._urlopener.version)
    self._say('Proxy: %s' % str(urllib._urlopener.proxies))

    if self.opts.simulate:
      self._say('Note : Simulation mode active')

    # Force options
    if self.opts.extract_all:
      self.opts.saveas = None

    # Batch mode
    if self.opts.batch:
      self._say('Batch: "%s"' % self.opts.batch)
      batch = readfilelines(self.opts.batch)
    else:
      batch = [url]

    self._printsep()

    for url in batch:

      self._say('Get  : "%s"' % url)
      html = self._retrievedata(url)

      vurls = []
      if is_embed_url(url):
        # Assuming video extraction from a page with <embed> tags
        # Parse the page HTML of all <embed> tags
        vurls = self._parsepage(html)
      else:
        # Assuming video extraction from a video page
        # Moving on to extraction
        vurls.append(url)

      if len(vurls) == 0:
        raise CliveError('Error: Nothing to extract')

      for vurl in vurls:

        try:

          if vurl != url:
            # Get page (HTML) only if we haven't done so earlier
            self._say('Get  : "%s"' % vurl)
            html = self._retrievedata(vurl)

          saveas = self._parsefname(html)
          self._say('Save : "%s"' % saveas)

          if not self.opts.simulate:

            self._checkoverwrite(saveas)

            vurl = self._extraction_vurl(html, vurl)
            self._retrievedata(vurl, saveas)

        except CliveError,e:
          self._say(e.msg)

        if not self.opts.simulate:

          self._reencode(saveas)
          self._play(saveas)

        self._printsep()

        if not self.opts.extract_all:
          break

  def _retrievedata(self, url, saveas=None):
    """Wraps urllib.urlretrieve"""
    urlcleanup() # Clear cache
    (filename,headers) = urlretrieve(url, saveas, self._progress)
    self._say('\n')
    if self.opts.http_debug:
      self._say('-'*70)
      v = self.opts.strip
      self.opts.strip = False # Disable strip temporarily
      self._say('%s' % str(headers))
      self.opts.strip = v
      self._say('-'*70)
    if not saveas:
      return readfile(filename)

  def _parsepage(self, html):
    """Parses HTML for any <embed> tags"""
    data = StringIO(html)
    parser = EmbedParser()
    parser.feed(data.read())
    parser.close()
    return parser.urls

  def _parsefname(self, html):
    """Returns output filename for the extracted video"""
    if self.opts.saveas:
      return self.opts.saveas
    data = StringIO(html)
    fmt = formatter.AbstractFormatter(formatter.NullWriter())
    parser = HTMLParser(fmt)
    parser.feed(data.read())
    parser.close()
    return parse_fname(parser.title)

  def _extraction_vurl(self, html, url):
    """Returns the video extraction URL (host specific)"""
    if url.find('youtube.com') != -1:
      p = YTubeParser()
      return p.parse(html)
    elif url.find('video.google.com') != -1:
      p = VGoogleParser()
      return p.parse(html, url, self._progress)
    else:
      (host,path) = parse_url(url)
      raise CliveError('Error: Invalid host (%s)' % host)

  def _checkoverwrite(self, fn):
    """Checks if file exists, prompts for confirmation if needed"""
    o = self.opts
    if os.path.exists(fn) and not o.overwrite and o.verbose:
      a = self._onprompt('Input: Overwrite file? (y/N):')
      if a.lower() != 'y':
        raise CliveError('Note : Keep %s' % fn)

  def _reencode(self, saveas, format=None):
    """Re-encodes extracted video to another format with ffmpeg"""
    if not self.opts.ffmpeg:
      return

    if not format:
      if self.opts.mpeg: self._reencode(saveas, 'mpeg')
      if self.opts.avi: self._reencode(saveas, 'avi')
      if self.opts.flv: self._reencode(saveas, 'flv')

    else:
      if saveas.endswith('.'+format):
        self._say('Note : Ignoring --%s, format exists already' % format)
        return

      # Check for executable
      (path_exe,args) = self._checkbinary(self.opts.ffmpeg)
      (ffmt,fout,ferr) = self._stdfilenames(saveas,format)

      self._say('Note : %s -> %s (%s)' % (saveas,format,path_exe))
      cmd = '"%s" %s -y -i "%s" "%s"' % (path_exe, args, saveas, ffmt)
      (fin,fout,ferr) = os.popen3(cmd)

      self._checkffmpeg(ferr)

  def _play(self, saveas):
    """Play video with the specified player"""
    if self.opts.play:
      (path_exe,args) = self._checkbinary(self.opts.player)

      self._say('Play : %s (%s)' % (saveas,path_exe))
      (ffmt,fout,ferr) = self._stdfilenames(saveas)

      cmd = '"%s" %s "%s"' % (path_exe, args, saveas)
      (fin,fout,ferr) = os.popen3(cmd)

      self._checkplayer(fout,ferr)

  def _checkbinary(self, path):
    """
    Checks if binary executable exists.
    Look from given path, if that fails, try the paths
    in the PATH environment variable.
    """
    (path,exe,args) = self._splitpath(path)
    path_exe = os.path.join(path,exe)

    if not os.path.isfile(path_exe):
      if not self._checkenvpath(path_exe):
        raise CliveError('Error: %s not found' % path_exe)

    return (path_exe,args)

  def _checkenvpath(self,bin):
    """Look from the environment variable PATHs"""
    for p in os.environ['PATH'].split(os.pathsep):
      n = os.path.join(p,bin)
      if sys.platform == 'win32':
        if not n.lower().endswith('.exe'):
          n += '.exe'
      if os.path.isfile(n):
        return True
    return False

  def _splitpath(self, path):
    """Splits path to path, executable and arguments"""
    path,exe = os.path.split(path)
    exe = exe.split(None,1)

    args = ''
    if len(exe) > 1:
      args = exe[1]

    return (path,exe[0],args)

  def _stdfilenames(self, saveas, format=None):
    """Returns <saveas>[.<fmt>.stdout,.stderr] filenames"""
    if format:
      ffmt = '%s.%s' % (saveas,format)
      fout = ffmt + '.stdout'
      ferr = ffmt + '.stderr'
    else:
      ffmt = None
      fout = saveas + '.stdout'
      ferr = saveas + '.stderr'
    return (ffmt,fout,ferr)

  def _checkffmpeg(self, ferr):
    self._checkstream(ferr,'stderr')

  def _checkplayer(self, fout, ferr):
    self._checkstream(fout,'stdout')
    self._checkstream(ferr,'stderr')

  def _checkstream(self, stream, name):
    """Checks the streams returned by popen3()"""
    data = stream.read()

    # No data or --quiet used
    if len(data) == 0 or not self.opts.verbose:
      return False

    ld = data.lower()

    if ld.find('access_file access error') != -1:      
      # vlc prints the above error frequently, and as far as the
      # playing goes, they seem harmless and are more of an annoyance.

      # If you need to see them, comment the 'if' line above
      # and the line below
      return False

    if ld.find('error') != -1:
      # Safe to assume if 'error' appears in the stdout/stderr
      # an error must have occurred
      a = self._onprompt('Error: Show %s? (Y/n):' % name)
      if not a.lower() == 'n':
        self._say('-- Begins.')
        v = self.opts.strip
        self.opts.strip = False # Disable strip temporarily
        self._say(data)
        self.opts.strip = v
        self._say('-- Ends.')
        return True

    return False


class EmbedParser(HTMLParser):
  """Parses HTML for <embed> tags with video URLs"""
  def __init__(self):
    self.urls = []
    fmt = formatter.AbstractFormatter(formatter.NullWriter())
    HTMLParser.__init__(self,fmt)

  def unknown_starttag(self, tag, attr):
    if tag.lower().find('embed') != -1:
      for a in attr:
        if len(a) > 1:
          if a[0] == 'src':
            self._appendurl(a[1])

  # Non-public

  def _appendurl(self, url):
    """Append URL to list if it qualifies"""

    if url.startswith('/'):
      # Treat URL as relative - insert host
      if self.title.find('YouTube -') != -1:
        url = 'http://www.youtube.com' + url
      elif self.title.find('- Google Video') != -1:
        url = 'http://video.google.com' + url
      else:
        raise CliveError('Error: Unknown host (tried to id from title)')

    # Convert embed to video page URL (if app.)
    url = normalize_url(url)

    for u in self.urls:
      # Look for a duplicate
      if u == url: return

    self.urls.append(url)


class YTubeParser:
  """Video extraction URL constructor for YouTube"""
  def parse(self, html):
    """
    Constructs video extraction URL.

    YouTube has a unique ID for each video on the website.
    These IDs can be extracted from the video page HTML,
    and be used with the 'get_video' (see below) to extract
    the actual flash video from the website.
    """
    text = parse_fromto(html,'player2.swf','"')

    if len(text) == 0:
      # Check if the page is flagged as 'mature content' and
      # terminate. Clive cannot handle login/cookies. Yet.
      if html.find('please verify you are 18 or older') != -1:
        raise CliveError('Error: Terminated (reason: age-verification)')
      else:
        raise CliveError('Error: Video ID not found in page HTML')

    return 'http://www.youtube.com/get_video?video_id=' + \
            text.split('video_id=')[1]


class VGoogleParser:
  """Video extraction URL constructor for Google Video"""
  def parse(self, html, url, onprogress):
    """
    Constructs video extraction URL.

    Google Video stores video details in gvp files,
    including the video extraction link. The URL to the
    gvp file can be found in the video page HTML.
    """
    if url.find('/videogvp/') == -1:
      # Not a gvp URL, extract gvp URL from HTML
      url = self._parsegvpurl(html)

    return self._parsegvp(url, onprogress)

  # Non-public

  def _parsegvpurl(self, html):
    """Extracts video gvp URL from HTML"""
    text = parse_fromto(html,'http://video.google.com/videogvp/','&')

    if len(text) == 0:
      raise CliveError('Error: Video gvp URL not found in page HTML')

    return text

  def _parsegvp(self, url, onprogress):
    """Downloads video gvp file and parses it for video extraction URL"""
    (filename,headers) = urlretrieve(url, None, onprogress)
    return parse_gvp(readfile(filename), 'url:')
