#############################################################################
# clive, video extraction utility
# Copyright (C) 2007 Toni Gundogdu
#
# clive is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  0.1.2-1307  USA
#############################################################################

import os
import ConfigParser

from optparse import OptionParser

from clive.verinfo import *


__all__ = [
  'Options'
]


class Options:
  """Parses both runtime configuration and command line options.
     NOTE: latter overrides"""
  CLIVE_CONF_FILE = '.cliverc'
  CLIVE_CONF_SECTION = 'clive'

  def __init__(self):
    self.cfile = os.path.join(os.path.expanduser('~'), self.CLIVE_CONF_FILE)

  def parse(self):
    conf = self._parseconf()
    (opts,args) = self._parseopts(conf)
    return (opts,args,conf)

  # Non-public

  def _parseconf(self):
    """Parses runtime configuration"""
    conf = ConfigParser.SafeConfigParser()
    conf.read(self.cfile)
    return conf

  def _parseopts(self, conf):
    """Parses cmdline options, overrides runtimeconf, saves if requested"""
    p = OptionParser(
      usage = PACKAGE_USAGE,
      description = PACKAGE_DESC,
      version = '%prog '+PACKAGE_VERSION+'\n'+PACKAGE_NOTICE
    )

    (n,v) = (PACKAGE_NAME,PACKAGE_VERSION)

    p.add_option('--agent',
      dest='agent',
      default=None,
      help='user-agent for http requests, e.g. --agent="%s/%s"' % (n,v))
    p.add_option('--proxy',
      dest='proxy',
      default={},
      help='http proxy for connections (possible values: [url|env])')
    p.add_option('--player',
      dest='player',
      default=None,
      metavar='PATH',
      help='path to player, e.g. --player="/usr/bin/ffplay"')
    p.add_option('--ffmpeg',
      dest='ffmpeg',
      default=None,
      metavar='PATH',
      help='path to ffmpeg, e.g. --ffmpeg="/usr/bin/ffmpeg"')
    p.add_option('--save',
      dest='save',
      action='store_true',
      default=False,
      help='save above runtime options to runtime configuration')
    p.add_option('--batch',
      dest='batch',
      default=None,
      metavar='PATH',
      help='import urls from a batch file (text:url/line) [ignores URL]')
    p.add_option('--extract-all',
      dest='extract_all',
      action='store_true',
      default=False,
      help='extract all embedded videos from url ' + '(disables: --saveas)')
    p.add_option('--simulate',
      dest='simulate',
      action='store_true',
      default=False,
      help="simulate extraction (scan, print but don't extract videos)")
    p.add_option('--saveas',
      dest='saveas',
      default=None,
      metavar='PATH',
      help='save extracted video as (see also --extract-all)')
    p.add_option('--overwrite',
      dest='overwrite',
      action='store_true',
      default=False,
      help='overwrite existing file')
    p.add_option('--mpeg',
      dest='mpeg',
      action='store_true',
      default=False,
      help='re-encode extracted to mpeg (see also --ffmpeg)')
    p.add_option('--avi',
      dest='avi',
      action='store_true',
      default=False,
      help='re-encode extracted video to avi (see also --ffmpeg)')
    p.add_option('--flv',
      dest='flv',
      action='store_true',
      default=False,
      help='re-encode extracted video to flv (see also --ffmpeg)')
    p.add_option('--play',
      dest='play',
      action='store_true',
      default=False,
      help='play video after extraction (see also --player)')
    p.add_option('--quiet',
      dest='verbose',
      action='store_false',
      default=True,
      help="don't print status messages to stdout (forces --overwrite)")
    p.add_option('--nostrip',
      dest='strip',
      action='store_false',
      default=True,
      help="don't strip printed stdout status messages")
    p.add_option('--http-debug',
      dest='http_debug',
      action='store_true',
      default=False,
      help='print http debug messages to stdout')

    (opts,args) = p.parse_args()

    self._overrideconf(opts,conf)
    self._saveconf(opts,conf)

    return (opts,args)

  def _overrideconf(self, opts, conf):
    if not opts.agent:
      opts.agent = self._getconf(conf, 'agent')
    if not opts.proxy:
      opts.proxy = self._getconf(conf, 'proxy')
    if not opts.player:
      opts.player = self._getconf(conf, 'player')
    if not opts.ffmpeg:
      opts.ffmpeg = self._getconf(conf, 'ffmpeg')

  def _saveconf(self, opts, conf):
    if opts.save:
      self._setconf(conf, 'agent', opts.agent)
      self._setconf(conf, 'proxy', opts.proxy)
      self._setconf(conf, 'player', opts.player)
      self._setconf(conf, 'ffmpeg', opts.ffmpeg)
      conf.write(open(self.cfile,'w'))

  def _getconf(self, conf, name):
    try:
      return conf.get(self.CLIVE_CONF_SECTION, name)
    except:
      return None

  def _setconf(self, conf, name, value):
    if not value:
      return

    try:
      conf.set(self.CLIVE_CONF_SECTION, name, value)
    except ConfigParser.NoSectionError:
      conf.add_section(self.CLIVE_CONF_SECTION)
      self._setconf(conf,name,value)
    except:
      pass
