#############################################################################
# clive, video extraction utility
# Copyright (C) 2007 Toni Gundogdu
#
# clive is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  0.1.2-1307  USA
#############################################################################

import re
import random
import string
import os.path

from urlparse import *
from urllib import *


__all__ = [
  # Functions
  'normalize_url',
  'is_embed_url',
  'parse_fname',
  'parse_url',
  'parse_fromto',
  'parse_gvp',
  'readfile',
  'readfilelines',
  # Classes
  'CliveOpener',
  'CliveError'
]


def normalize_url(url):
  """
  'Normalizes' the given URL. Converts parts of the URL to lower case,
  converts an embed URL to a video page URL and inserts http:// if
  needed.
  """

  # Convert parts of URL to lower case
  # NOTE: path stays the same
  scheme,host,path,params,query,fragment = urlparse(url)
  scheme = scheme.lower()
  host = host.lower()

  if host.find('video.google.com') != -1:
    # Google: 'docid' must be lower-case (some websites neglect this)
    # YouTube: video IDs are case-sensitive, query remains as it is
    query = query.lower()

  url = urlunparse((scheme,host,path,params,query,fragment))

  # Embed -> video page
  url = embed_to_vpageurl(url)

  # Insert http:// if needed
  if not url.startswith('http://'):
    url = 'http://' + url

  return url


def embed_to_vpageurl(url):
  """Converts an embed video URL to a video page URL"""
  if url.lower().find('youtube.com') != -1:
    # YouTube specific
    url = url.replace('/v/','/watch?v=')
    url = url.replace('http://youtube.com','http://www.youtube.com')
  elif url.lower().find('video.google.com') != -1:
    # Google Video specific
    url = url.replace('/googleplayer.swf?','/videoplay?')
  return url


def is_embed_url(url):
  """Returns True if the specified URL is an embed video URL"""
  r = True
  if url.find('/watch?v=') != -1:
    r = False
  elif url.find('/videoplay?') != -1:
    r = False
  return r


def parse_fname(title):
  """
  Tries to use <title/> for output filename, otherwise
  a random string is used for filename. Appends also file
  extension to a filename, based on the host.
  """
  ext = ''

  if title.find('YouTube -') != -1:
    ext = '.flv'
  else:
    ext = '.avi'

  title = title.replace('YouTube -','')
  title = title.replace('- Google Video','')

  title = re.sub('[^A-Za-z0-9]','',title)

  if len(title) == 0:
    title = ''.join([random.choice(string.letters) for c in xrange(8)])

  return title + ext


def parse_url(url):
  """Parses URL, returns host,path, inserts 'http://' if needed"""

  # NOTE: Python >= 2.5 offers less cumbersome way to do this
  scheme,host,path,params,query,fragment = urlparse(url)

  # Reconstruct the path with path+query
  path = urlunparse((None,None,path,None,query,None))

  return (host,path)


def parse_fromto(data, _from, _to,skip_from=False):
  """Basic text parser, extracts text between from-to"""

  start = data.find(_from)
  end = data.find(_to,start)

  text = ''

  if start != -1 and end != -1:
    if skip_from: start += len(_from)
    text = data[start:end]

  return text


def parse_gvp(data, find):
  """Basic Google Video GVP file parser"""
  return parse_fromto(data,find,'\n',skip_from=True)


def readfile(fn):
  """Reads a file into memory and returns the data"""
  if not os.path.exists(fn):
    raise CliveError('Error: %s not found' % fn)
  f = open(fn,'r')
  data = f.read()
  f.close()
  return data


def readfilelines(fn):
  """Reads a file into memory and returns a list of lines"""
  if not os.path.exists(fn):
    raise CliveError('Error: %s not found' % fn)
  f = open(fn,'r')
  lines = []
  for ln in f.xreadlines():
    ln = ln.rstrip('\r\n')
    lines.append(ln)
  f.close()
  return lines


class CliveOpener(FancyURLopener):
  """Custom URL opener, sets custom user-agent for connections"""
  def __init__(self, agent, proxies):
    if agent:
      self.version = agent
    FancyURLopener.__init__(self, proxies)


class CliveError(Exception):
  """Clive error exception"""
  def __init__(self,msg): self.msg = msg
